#include <_Custom_MetaInfo/Ativo.mqh>
#include <Trade/PositionInfo.mqh>
#include <Trade/AccountInfo.mqh>

class Operacoes {
private:
    Ativo* ativo;
    CPositionInfo* posicao;
    CAccountInfo* conta;
    double volume;
    ulong magic;

    double stopLoss;
    double takeProfit;

public:
    Operacoes(){ this.set(); };
    ~Operacoes(){};

    //Define o padrão de envio de ordens
    void set();

    //Define o padrão de envio de ordens com paramêtros
    void setOrderDefault(string simbolo_, double volume_, ulong magic_);

    //Define o padrão de envio de ordens com paramêtros
    void setDefaultStops(double takeProfit_, double stopLoss_);

    //Envio de ordens a mercado
    bool operacaoAMercado(string simbolo_, ENUM_ORDER_TYPE tipoOrdem_, double volume_, double preco_, ulong magic_);

    bool envioStops(string simbolo_, ulong magic_, double takeProfit_, double stopLoss_);

    //Realiza tentantiva de envio de ordem
    bool tentarEnviarOrdem(MqlTradeRequest& requisicao_, MqlTradeResult& resultado_, int tempoEntreTentativas, int maxTentativas_);

    //Realização de compra padrão, a mercado
    bool compraPadrao();
    //Realização de venda padrão, a mercado
    bool vendaPadrao();
    //Envio de stop padrão
    bool stopPadrao();
    bool posicaoAberta(){ return posicao.Select(this.ativo.getSimbolo()); };
};

void Operacoes::set(){
    this.setOrderDefault(_Symbol, 1, 111324);
    this.setDefaultStops(20, 15);
    this.posicao = new CPositionInfo;
    this.conta = new CAccountInfo;
}

void Operacoes::setOrderDefault(string simbolo_, double volume_, ulong magic_){
    this.ativo = new Ativo(simbolo_);
    this.volume = this.ativo.getVolume(volume_);
    this.magic = magic_;
}

void Operacoes::setDefaultStops(double takeProfit_, double stopLoss_){
    this.stopLoss = stopLoss_;
    this.takeProfit = takeProfit_;
}

bool Operacoes::operacaoAMercado(string simbolo_, ENUM_ORDER_TYPE tipoOrdem_, double volume_, double preco_, ulong magic_){
    MqlTradeRequest requisicao = {0};
    MqlTradeResult resultado = {0};

    requisicao.action = TRADE_ACTION_DEAL;
    requisicao.symbol = simbolo_;
    requisicao.volume = volume_;
    requisicao.magic = magic_;
    requisicao.type = tipoOrdem_;
    requisicao.price = preco_;
    requisicao.deviation = 5;

    double margem;
    double margemLivre = conta.FreeMargin();

    if(!OrderCalcMargin(tipoOrdem_, simbolo_, volume_, preco_, margem)){
      Print("Erro na validação");
      return false;
    }
    
    if(margem > margemLivre){
        Print("Margem insuficiente!");
        return false;
    }

    if(!tentarEnviarOrdem(requisicao, resultado, 2, 3)){
        Print("Falha na tentativa de operação a mercado.");
        return false;
    }
    return true;
}

bool Operacoes::envioStops(string simbolo_, ulong magic_, double takeProfit_, double stopLoss_){
    if(!this.posicaoAberta())
        return false;

    MqlTradeRequest requisicao = {0};
    MqlTradeResult resultado = {0};
    //Trata conforme os pontos do ativo o stop loss e o take profit 
    double tkTratado = ativo.getPontos(takeProfit_);
    double slTratado = ativo.getPontos(stopLoss_);
    double tp = 0;
    double sl = 0;
    //Agrega informações da posição
    ENUM_POSITION_TYPE tipoPosicao = posicao.PositionType();
    double precoAtual = posicao.PriceOpen();
    double slAtual = NormalizeDouble(posicao.StopLoss(), ativo.getDigitos());
    double tpAtual = NormalizeDouble(posicao.TakeProfit(), ativo.getDigitos());

    requisicao.action = TRADE_ACTION_SLTP;
    requisicao.position = posicao.Ticket();
    requisicao.symbol = simbolo_;
    requisicao.magic = magic_;

    if(tipoPosicao == POSITION_TYPE_BUY){
        tp = precoAtual + tkTratado;
        sl = precoAtual - slTratado;
        if(tp > tpAtual && tpAtual != 0)
            tp = tpAtual;
        if(sl < slAtual && slAtual != 0)
            sl = slAtual;
    }else if(tipoPosicao == POSITION_TYPE_SELL){
        tp = precoAtual - tkTratado;
        sl = precoAtual + slTratado;
        if(tp < tpAtual && tpAtual != 0)
            tp = tpAtual;
        if(sl > slAtual && slAtual != 0)
            sl = slAtual;
    }

    requisicao.sl = NormalizeDouble(sl, ativo.getDigitos());
    requisicao.tp = NormalizeDouble(tp, ativo.getDigitos());
    

    if((sl > 0 && sl != slAtual) || (tp > 0 && tp != tpAtual)){
        return tentarEnviarOrdem(requisicao, resultado, 2, 3);
    }
    return true;
}

bool Operacoes::tentarEnviarOrdem(MqlTradeRequest& requisicao_, MqlTradeResult& resultado_, int tempoEntreTentativas, int maxTentativas_){
    ResetLastError();
    int maxTentativas = maxTentativas_;
    int numTentativas = 0;
    string mensagens = "";
    bool sucesso = false;

    while(numTentativas < maxTentativas){
        //Soma no número de tentativas
        numTentativas++;
        //Envia a requisição
        sucesso=OrderSend(requisicao_,resultado_);
        //Se o envio falhar, valida o motivo e faz retentativa
        if(!sucesso){
            uint codRetorno = resultado_.retcode;
            mensagens += "Requisição com erro ";
            switch(codRetorno){
                //--- requote
                case 10004:{
                    mensagens += "TRADE_RETCODE_REQUOTE | ";
                    break;
                }
                //--- order is not accepted by the server
                case 10006:{
                    mensagens += "TRADE_RETCODE_REJECT | ";
                    break;
                }
                //--- invalid price
                case 10015:{
                    mensagens += "TRADE_RETCODE_INVALID_PRICE | ";
                    break;
                }
                //--- invalid SL and/or TP
                case 10016:{
                    mensagens += "TRADE_RETCODE_INVALID_STOPS | ";
                    break;
                }
                //--- invalid volume
                case 10014:{
                    mensagens += "TRADE_RETCODE_INVALID_VOLUME | ";
                    break;
                }
                //--- not enough money for a trade operation 
                case 10019:{
                    mensagens += "TRADE_RETCODE_NO_MONEY | ";
                    break;
                }
                //--- some other reason, output the server response code 
                default:{
                    break;
                }
            }
            mensagens += "Código de erro = "+IntegerToString(codRetorno)+"\n";
            Sleep(tempoEntreTentativas*1000);
        }else{
            mensagens += "Ordem enviada com sucesso em "+IntegerToString(numTentativas)+" tentativas.";
            break;
        }
    }
    Print(mensagens);
    //Retorna se houve sucesso na operação
    return sucesso;
}

bool Operacoes::compraPadrao(){
    return this.operacaoAMercado(this.ativo.getSimbolo(), ORDER_TYPE_BUY, this.volume, this.ativo.getLast(), this.magic);
}

bool Operacoes::vendaPadrao(){
    return this.operacaoAMercado(this.ativo.getSimbolo(), ORDER_TYPE_SELL, this.volume, this.ativo.getLast(), this.magic);
}

bool Operacoes::stopPadrao(){
    return this.envioStops(this.ativo.getSimbolo(), this.magic, this.takeProfit, this.stopLoss);
}